use common::Part;
use regex::Regex;
use std::fmt::Display;

use thiserror::Error;

#[derive(Error, Debug)]
pub enum Day01Error {
    #[error("something weird happened")]
    RuntimeError,
}

pub fn run_part(part: Part, inp: String) -> Result<Box<dyn Display>, Day01Error> {
    match part {
        Part::One => Ok(Box::new(part_01(&inp))),
        Part::Two => Ok(Box::new(part_02(&inp))),
    }
}

fn part_01(inp: &str) -> usize {
    let split_inp = inp.lines();

    let mut sum = 0;
    for line in split_inp {
        let matches: Vec<&str> = line.matches(char::is_numeric).collect();
        let (first, last) = (matches.first().unwrap(), matches.last().unwrap());
        sum += format!("{}{}", first, last).parse::<usize>().unwrap();
    }

    sum
}

fn get_nums(inp: &str) -> Vec<&str> {
    let re =
        Regex::new(r"(one|two|three|four|five|six|seven|eight|nine|1|2|3|4|5|6|7|8|9)").unwrap();
    return re.find_iter(inp).map(|m| m.as_str()).collect();
}

fn convert_num(inp: &str) -> usize {
    match inp {
        "1" | "one" => 1,
        "2" | "two" => 2,
        "3" | "three" => 3,
        "4" | "four" => 4,
        "5" | "five" => 5,
        "6" | "six" => 6,
        "7" | "seven" => 7,
        "8" | "eight" => 8,
        "9" | "nine" => 9,
        _ => unimplemented!(),
    }
}

fn part_02(inp: &str) -> usize {
    let split_inp = inp.lines().map(|line| get_nums(line)).collect::<Vec<_>>();

    let mut sum: usize = 0;
    for line in split_inp {
        let (first, last) = (
            convert_num(line.first().unwrap()),
            convert_num(line.last().unwrap()),
        );
        sum += format!("{}{}", first, last).parse::<usize>().unwrap()
    }
    sum
}

#[cfg(test)]
mod part_01_tests {
    use crate::part_01;

    #[test]
    fn test_line() {
        assert_eq!(part_01("1abc2"), 12)
    }

    #[test]
    fn test_lines() {
        assert_eq!(
            part_01(
                r#"1abc2
pqr3stu8vwx
a1b2c3d4e5f
treb7uchet"#
            ),
            142
        )
    }
}

#[cfg(test)]
mod part_02_tests {
    use crate::{get_nums, part_02};

    #[test]
    fn test_regex() {
        assert_eq!(get_nums("two1nine"), vec!["two", "1", "nine"])
    }

    #[test]
    fn test_line() {
        assert_eq!(part_02("two1nine"), 29)
    }

    #[test]
    fn test_lines() {
        assert_eq!(
            part_02(
                r#"two1nine
eightwothree
abcone2threexyz
xtwone3four
4nineeightseven2
zoneight234
7pqrstsixteen"#
            ),
            281
        )
    }

    #[test]
    fn test_more_lines() {
        assert_eq!(
            part_02(
                r#"ckmb52fldxkseven3fkjgcbzmnr7
gckhqpb6twoqnjxqplthree2fourkspnsnzxlz1
2onetwocrgbqm7
frkh2nineqmqxrvdsevenfive
four9two
six7sixqrdfive3twonehsk
xkvsone2
one65
rggxsff1seven
djbcgrrtqdshpqqzj43rgcr"#
            ),
            361
        )
    }
}
