use std::fmt::Display;
use std::mem;
use std::str::FromStr;

use thiserror::Error;
use winnow::ascii::digit1;
use winnow::combinator::{alt, preceded, separated, terminated};
use winnow::error::InputError;
use winnow::prelude::*;
use winnow::PResult;

use common::Part;

use crate::Color::{Blue, Green, Red};

#[derive(Error, Debug)]
pub enum Day02Error {
    #[error("something weird happened")]
    RuntimeError,

    #[error("Couldn't parse game")]
    GameParseError,
}

#[derive(Debug, PartialEq, Copy, Clone)]
enum Color {
    Green(usize),
    Red(usize),
    Blue(usize),
}

#[derive(Debug, Default, PartialOrd, PartialEq, Copy, Clone)]
struct Reveal {
    green: usize,
    red: usize,
    blue: usize,
}

#[derive(Debug, Default, PartialOrd, PartialEq)]
struct Game {
    id: usize,
    reveals: Vec<Reveal>,
}

impl Game {
    fn fits_requirements(&self, req: &Requirements) -> bool {
        let max_red = self
            .reveals
            .iter()
            .max_by(|&x, &y| x.red.cmp(&y.red))
            .expect("No max red")
            .red;
        let max_green = self
            .reveals
            .iter()
            .max_by(|&x, &y| x.green.cmp(&y.green))
            .expect("No max green")
            .green;
        let max_blue = self
            .reveals
            .iter()
            .max_by(|&x, &y| x.blue.cmp(&y.blue))
            .expect("No max blue")
            .blue;

        max_red <= req.red && max_blue <= req.blue && max_green <= req.green
    }
}

#[derive(Debug, Default)]
struct Requirements {
    red: usize,
    green: usize,
    blue: usize,
}

fn parser_blue<'s>(input: &mut &'s str) -> PResult<Color, InputError<&'s str>> {
    terminated(digit1, " blue")
        .map(|s: &str| Color::Blue(s.parse().unwrap()))
        .parse_next(input)
}

fn parse_blue(input: &mut &str) -> Color {
    parser_blue(input).unwrap()
}

fn parser_red<'s>(input: &mut &'s str) -> PResult<Color, InputError<&'s str>> {
    terminated(digit1, " red")
        .map(|s: &str| Color::Red(s.parse().unwrap()))
        .parse_next(input)
}

fn parse_red(input: &mut &str) -> Color {
    parser_red(input).unwrap()
}

fn parser_green<'s>(input: &mut &'s str) -> PResult<Color, InputError<&'s str>> {
    terminated(digit1, " green")
        .map(|s: &str| Color::Green(s.parse().unwrap()))
        .parse_next(input)
}

fn parse_green(input: &mut &str) -> Color {
    parser_green(input).unwrap()
}

fn parse_colors<'s>(input: &mut &'s str) -> PResult<Vec<Color>, InputError<&'s str>> {
    separated(0.., alt((parser_green, parser_blue, parser_red)), ", ").parse_next(input)
}

fn colors_into_reveal(colors: Vec<Color>) -> Reveal {
    let Green(green) = colors
        .iter()
        .find(|&a| mem::discriminant(a) == mem::discriminant(&Color::Green(0)))
        .unwrap_or(&Color::Green(0))
    else {
        panic!()
    };

    let Blue(blue) = colors
        .iter()
        .find(|&a| mem::discriminant(a) == mem::discriminant(&Color::Blue(0)))
        .unwrap_or(&Color::Blue(0))
    else {
        panic!()
    };

    let Red(red) = colors
        .iter()
        .find(|&a| mem::discriminant(a) == mem::discriminant(&Color::Red(0)))
        .unwrap_or(&Color::Red(0))
    else {
        panic!()
    };

    Reveal {
        red: *red,
        blue: *blue,
        green: *green,
    }
}
fn parse_reveal<'a>(inp: &mut &'a str) -> PResult<Reveal, InputError<&'a str>> {
    parse_colors.map(colors_into_reveal).parse_next(inp)
}

// fn parse_reveals(inp: &mut &str) -> Vec<Reveal> {
//     separated(0.., )
// }

fn parse_reveals<'a>(inp: &mut &'a str) -> PResult<Vec<Reveal>, InputError<&'a str>> {
    separated(0.., parse_reveal, "; ").parse_next(inp).clone()
    // .map(parse_reveal)
    // .parse_next(&mut input)?)
    // separated(0.., parse_colors, "; ").map(|s: &str| parse_reveal(s)).parse_next(&mut input)
}

fn parser_game<'s>(input: &mut &'s str) -> Game {
    let game_num = preceded("Game ", digit1::<&'s str, InputError<&'s str>>)
        .parse_next(input)
        .unwrap();
    let colors = preceded(": ", parse_reveals).parse_next(input).unwrap();
    dbg!(&game_num, &colors);

    Game {
        id: game_num.parse().unwrap(),
        reveals: colors.clone(),
    }
}

impl TryFrom<&str> for Game {
    type Error = Day02Error;

    fn try_from(value: &str) -> Result<Self, Self::Error> {
        value.parse()
    }
}

impl FromStr for Game {
    type Err = Day02Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let mut s = s;
        Ok(parser_game(&mut s))
    }
}

pub fn run_part(part: Part, inp: String) -> Result<Box<dyn Display>, Day02Error> {
    match part {
        Part::One => Ok(Box::new(part_01(&inp))),
        Part::Two => Ok(Box::new(part_02(&inp))),
    }
}

fn part_01(inp: &str) -> usize {
    let parsed_input: Vec<Game> = inp.lines().map(|x| x.try_into().unwrap()).collect();
    let requirements = Requirements {
        red: 12,
        green: 13,
        blue: 14,
    };
    parsed_input
        .into_iter()
        .filter(|game| game.fits_requirements(&requirements))
        .fold(0, |acc, game| acc + game.id)
}

fn part_02(inp: &str) -> usize {
    todo!()
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_parse_blue() {
        assert_eq!(parse_blue(&mut "5 blue"), Color::Blue(5))
    }

    #[test]
    fn test_parse_reveal() {
        assert_eq!(
            parse_reveal(&mut "3 blue, 4 red").unwrap(),
            Reveal {
                green: 0,
                blue: 3,
                red: 4,
            }
        )
    }

    #[test]
    fn test_parse_game() {
        assert_eq!(
            "Game 1: 3 blue, 4 red; 1 red, 2 green, 6 blue; 2 green"
                .parse::<Game>()
                .unwrap(),
            Game {
                id: 1,
                reveals: vec![
                    Reveal {
                        blue: 3,
                        red: 4,
                        green: 0,
                    },
                    Reveal {
                        red: 1,
                        green: 2,
                        blue: 6,
                    },
                    Reveal {
                        green: 2,
                        blue: 0,
                        red: 0,
                    },
                ],
            }
        )
    }

    #[test]
    fn test_meets_reqs() {
        assert!(Game {
            id: 1,
            reveals: vec![
                Reveal {
                    blue: 3,
                    red: 4,
                    green: 0,
                },
                Reveal {
                    red: 1,
                    green: 2,
                    blue: 6,
                },
                Reveal {
                    green: 2,
                    blue: 0,
                    red: 0,
                },
            ],
        }
        .fits_requirements(&Requirements {
            red: 10,
            blue: 10,
            green: 10,
        }))
    }

    #[test]
    fn test_does_not_meet_reqs() {
        assert!(
            !(Game {
                id: 1,
                reveals: vec![
                    Reveal {
                        blue: 3,
                        red: 4,
                        green: 0,
                    },
                    Reveal {
                        red: 1,
                        green: 2,
                        blue: 6,
                    },
                    Reveal {
                        green: 2,
                        blue: 0,
                        red: 0,
                    },
                ],
            }
            .fits_requirements(&Requirements {
                red: 5,
                blue: 5,
                green: 5,
            }))
        )
    }
}
